package pl.psnc.dlibramuseum.app.extension.csv;

import java.awt.Graphics2D;
import java.awt.Window;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.StringTokenizer;

import pl.psnc.dlibra.app.common.TextWrapper;
import pl.psnc.dlibra.app.common.component.JSystemFileChooser;
import pl.psnc.dlibra.app.common.component.extension.AbstractConfigurableExtension;
import pl.psnc.dlibra.app.common.component.inprogress.OperationInProgressManager;
import pl.psnc.dlibra.app.extension.dictionarymanager.DictionaryManager;
import pl.psnc.dlibra.app.extension.dictionarymanager.DictionaryManagerException;

/**
 * Extension class from importing dictionary from files in quasi-CSV format.
 * 
 * @author margraf
 *
 *         <h3>(pl - from user documentation) Opis importowanego formatu</h3>
 *
 *         <ul>
 *         <li>Kodowanie pliku powinno być ustawione na UTF-8</li>
 *         <li>Każda linia pliku powinna być niezależnym wpisem, dotyczącym zestawu słów/fraz będących wyrazami z jednej
 *         grupy. Każda grupa to wyrazy bliskoznaczne lub synonimy.</li>
 *         <li>Każda linia powinna składać się z wartości bazowej (musi być na pierwszym miejscu) oraz 0 lub więcej
 *         wartości bliskoznacznych. Wartość bazowa i wartości bliskoznaczne muszą być oddzielone od siebie znakiem ";"</li>
 *         <li>Żadna wartość nie może zawierać znaku ";", gdyż jest to znak specjalny do rozdzielenia poszczególnych
 *         wartości.</li>
 *         </ul>
 * 
 */
public class CSVDictionaryManager extends AbstractConfigurableExtension implements DictionaryManager {

    /** Location of UI resources */ 
    private static final String DEFAULT_BUNDLE_PATH = "pl/psnc/dlibramuseum/app/extension/csv/res/GUI";
    
    /** Default charset of input file. There is possibility of parameterization in the future */
    private static final String DEFAULT_CHARSET = "utf8";
    
    /** Default delimiter in input file. There is possibility of parameterization in the future */
    private static final char DEFAULT_DELIMITER = ';';

    /** Pattern of progress message - read once per inport from resource */ 
    private String progressMessagePattern = null;
    
    /** Store current directory for file chooser */
    private static File currentDir = new File(".");


    /** Parameterless constructor */
    public CSVDictionaryManager() {
        reloadResourceBundle(DEFAULT_BUNDLE_PATH);
    }


    @Override
    protected void reloadResourceBundle(String bundlePath) {
        if (textsBundle == null || !Locale.getDefault().equals(textsBundle.getLocale())) {
            textsBundle = ResourceBundle.getBundle(bundlePath, Locale.getDefault());
        }
    }


    @Override
    public boolean isConfigurable() {
        return false;
    }


    @Override
    public void showConfiguration(Window parentWindow) {
        //not supported
    }


    @Override
    public Map<String, List<String>> loadDictionary(Window parentWindow)
            throws DictionaryManagerException {
        reloadResourceBundle(DEFAULT_BUNDLE_PATH);
        Map<String, List<String>> result = null;

        JSystemFileChooser systemFileChooser = new JSystemFileChooser();
        systemFileChooser.setCurrentDirectory(currentDir);
        
        if (systemFileChooser.showOpenDialog(parentWindow) == JSystemFileChooser.APPROVE_OPTION) {
            try {
                File selectedFile = systemFileChooser.getSelectedFile();
                int linesCounter = countLines(selectedFile);

                result = readParseFile(selectedFile, linesCounter, null);
            } catch (Exception e) {
                throw new DictionaryManagerException(TextWrapper.wrapText(
                    textsBundle.getString("CSVDictionaryManager.loadErrorMessage").replace("{1}", e.getMessage()),
                    parentWindow.getWidth() / 2, (Graphics2D) parentWindow.getGraphics()));
            }
        }
        
        //keep selected directory for next time
        currentDir = systemFileChooser.getCurrentDirectory();
        
        return result;
    }


    @Override
    public void storeDictionary(Window parentWindow, Map<String, List<String>> dictionary)
            throws DictionaryManagerException {
        //not supported
    }


    @Override
    public boolean supports(DictionaryAction action) {
        return DictionaryAction.LOAD.equals(action);
    }

    /**
     * Read given file, parse each line and put data from lines to result
     * 
     * @param selectedFile File to read data
     * @return Map with parsed data 
     * @throws IOException Throws if some I/O problem (e.g. no file) occurs 
     */
    /*package*/ Map<String, List<String>> readParseFile(File selectedFile, int linesInFile,
            OperationInProgressManager oipManager)
            throws IOException {
        FileInputStream fis = new FileInputStream(selectedFile);
        return readParseFile(fis, linesInFile, oipManager);
    }

    
    /**
     * Read data fro mgiven input stream, parse each line and put data from lines to result
     * 
     * @param inputStream Data input stream
     * @return Map with parsed data 
     * @throws IOException Throws if some I/O problem (e.g. no file) occurs 
     */
    /*package*/ Map<String, List<String>> readParseFile(InputStream inputStream, int linesInFile,
            OperationInProgressManager oipManager)
            throws IOException {
        reloadResourceBundle(DEFAULT_BUNDLE_PATH);
        
        Map<String, List<String>> result;

        LineNumberReader reader = null;
        String line = null;
        int currentLine = -1;

        try {
            reader = new LineNumberReader(new InputStreamReader(inputStream,  DEFAULT_CHARSET));
            result = new HashMap<String, List<String>>();
            progressMessagePattern = textsBundle.getString("CSVDictionaryManager.importing");

            line = reader.readLine();
            while (line != null) {
                if(oipManager != null) {
                    updateProgress(oipManager, currentLine++, linesInFile);
                }

                if (line.trim().length() > 0) {
                    StringTokenizer tokenizer = new StringTokenizer(line, String.valueOf(DEFAULT_DELIMITER));

                    List<String> items = new ArrayList<String>();
                    Set<String> testSet = new HashSet<String>();

                    String key = tokenizer.nextToken();
                    testSet.add(key);
                    
                    while (tokenizer.hasMoreTokens()) {
                        String token = tokenizer.nextToken();
                        
                        if(!testSet.contains(token)) {
                            items.add(token);
                            testSet.add(token);
                        }
                    }

                    result.put(key, items);
                }
                line = reader.readLine();
            }
        } finally {
            if (reader != null) {
                reader.close();
            }
        }

        return result;
    }


    /**
     * Simple method count lines in text file
     * 
     * @param file
     *            Text file to count lines
     * @return Number of lines in given file
     * @throws IOException
     *             throws if some problem with file IO operation occurs
     */
    private int countLines(File file)
            throws IOException {
        int result = 0;
        LineNumberReader reader = new LineNumberReader(new InputStreamReader(new FileInputStream(file),  DEFAULT_CHARSET));

        while (reader.readLine() != null) {
            result++;
        }

        reader.close();
        return result;
    }


    /**
     * Sets the current progress in oipManager object
     * 
     * @param oipManager
     *            Progress manager
     * @param current
     *            Current value of progress
     * @param maximum
     *            Maximum value of progress (required to formatting message text)
     */
    private void updateProgress(OperationInProgressManager oipManager, int current, int maximum) {
        //oipManager.initializeProgressBar(0, maximum);
        oipManager.updateProgressBar(current, false);
        oipManager.setProgressInfo(MessageFormat.format(progressMessagePattern, current + 1, maximum));
    }
    
    @Override
    public String toString() {
        reloadResourceBundle(DEFAULT_BUNDLE_PATH);
        return textsBundle.getString("CSVDictionaryManager.name");
    }


    

}
