package pl.psnc.dlibra.app.extension.html;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.activation.MimetypesFileTypeMap;

import org.cyberneko.html.parsers.DOMParser;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import pl.psnc.dlibra.app.extension.Extension;
import pl.psnc.dlibra.app.extension.fileparser.FileParser;
import pl.psnc.dlibra.app.parser.html.HtmlEncodingFinder;

/**
 * Parses HTML files. Build a list of files which are referenced from a given
 * file.
 */
public class HTMLFileParser
	implements FileParser, Extension
{

	private static final String HASH = "#";

	private static final String QUERY_CHAR = "?";

	private static final String SUPPORTED_MIMETYPE = "text/html";

	private static final String UNRECOGNIZED_MIMETYPE =  "application/octet-stream";

	private MimetypesFileTypeMap mimetypesFileTypeMap = new MimetypesFileTypeMap(
			this.getClass().getResourceAsStream(
				"/pl/psnc/dlibra/app/extension/html/file-types.properties"));

	private static final String[][] linkTags = { { "img", "src"},
			{ "a", "href"}, { "td", "background"}, { "frame", "src"},
			{ "body", "background"}, { "area", "href"}, { "link", "href"},
			{ "embed", "src"}};


	public List<URL> parseFile(URL file)
		throws Exception
	{
		if (file != null) {
			return buildFiles(file);
		}
		return null;
	}


	private List<URL> buildFiles(URL fileURL)
		throws IOException, SAXException
	{
		ArrayList<URL> referencedFiles = new ArrayList<URL>();

		String encoding = null;
		InputStream inputStream = null;
		try {
			inputStream = fileURL.openStream();
			encoding = new HtmlEncodingFinder().getEncoding(inputStream);
		}
		catch (Throwable e) {
			// do nothing
		}
		finally {
			try {
				inputStream.close();
			}
			catch (Exception e) {
				// do nothing
			}
		}

		DOMParser myParser = new DOMParser();
		InputSource is = new InputSource(fileURL.openStream());
		myParser.parse(is);

		for (String[] tagAttr : linkTags) {
			NodeList nodes = myParser.getDocument().getElementsByTagName(
				tagAttr[0]);

			for (int i = 0; i < nodes.getLength(); i++) {
				Node node = nodes.item(i);

				NamedNodeMap attributes = node.getAttributes();
				if (attributes == null) {
					continue;
				}

				Node namedItem = attributes.getNamedItem(tagAttr[1]);
				if (namedItem == null) {
					continue;
				}

				String href = namedItem.getNodeValue();

				if (href == null) {
					continue;
				}

				href = href.trim();

				try {
					href = URLDecoder.decode(href, encoding);
				}
				catch (Exception e) {
					// do nothing
				}

				if (pointsTheSameDocument(href) || isAbsoluteURI(href)
						|| isInvalid(href)) {
					continue;
				}

				if (href.indexOf(QUERY_CHAR) > -1) {
					href = href.substring(0, href.indexOf(QUERY_CHAR));
				}
				if (href.indexOf(HASH) > -1) {
					href = href.substring(0, href.indexOf(HASH));
				}

				URL link;
				try {
					link = fileURL.toURI().resolve(href).toURL();
				}
				catch (URISyntaxException e) {
					continue;
				}

				referencedFiles.add(link);
			}
		}

		return referencedFiles;
	}


	private boolean isInvalid(String href)
	{
		return "".equals(href);
	}


	private boolean isAbsoluteURI(String href)
	{
		return href.startsWith("/")
				|| href.toLowerCase().matches("[a-z]+[a-z0-9\\+\\-\\.]*:.+");
	}


	private boolean pointsTheSameDocument(String href)
	{
		return href.startsWith(HASH);
	}


	public boolean supports(URL file)
	{
		String contentType = mimetypesFileTypeMap.getContentType(file.getFile()
				.toLowerCase());
		if (contentType != null
				&& !UNRECOGNIZED_MIMETYPE.equalsIgnoreCase(contentType)) {
			return contentType.startsWith(SUPPORTED_MIMETYPE);
		}
		try {
			URLConnection connection = file.openConnection();
			if (connection instanceof HttpURLConnection) {
				HttpURLConnection urlConnection = (HttpURLConnection) file
						.openConnection();
				urlConnection.setRequestMethod("HEAD");
				if (urlConnection.getResponseCode() != HttpURLConnection.HTTP_OK) {
					throw new FileNotFoundException(file.toString());
				}
			}
			contentType = connection.getContentType();
			return contentType != null
					&& contentType.startsWith(SUPPORTED_MIMETYPE);
		}
		catch (IOException e) {
			return false;
		}
	}


	public void initialize(Map<String, Object> initPrefs)
	{
		// nothing to do
	}


	@Override
	public String toString()
	{
		return "HTML parser";
	}
}
