package pl.psnc.dlibra.app.extension.xml;


import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.XMLConstants;

import net.sf.saxon.javax.xml.xquery.XQException;
import net.sf.saxon.javax.xml.xquery.XQExpression;
import net.sf.saxon.javax.xml.xquery.XQResultSequence;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;


/**
 * XQuery used for XML conversion.
 * @version $Id: ConversionXQuery.java,v 1.3 2007-07-12 10:22:43 tparkola Exp $
 */
public class ConversionXQuery
{
	private static final String XML_LANG = "lang";
	private static final String DOCUMENT_INDICATOR = "{document}";
	private String query;

	/**
	 * Constructor. 
	 * @param query the XQuery to perform 
	 */
	public ConversionXQuery(String query) {
		this.query = query;
	}
	
	/**
	 * Executes query. The result contains values retrieved as query result
	 * in specific language (which is retrieved from the source file). If
	 * the source file has no information concerning the language of values then
	 * defaultLang is used.
	 * @param file the subject of the query (source file)
	 * @param expr expression which will be used to execute query
	 * @param defaultLang default language for values with no language specified 
	 * @return map which maps language to list of values, for example pl->(Tomasz, Jan), 
	 * en->(Thomas, John)
	 * @throws XQException 
	 */
	public Map<String, List<String>> executeQuery(File file, XQExpression expr, String defaultLang) 
	throws XQException {
		Map<String, List<String>> result = new HashMap<String, List<String>>();
		if (!file.exists()) {
			return result;
		}
		
		String q = this.query.replace(DOCUMENT_INDICATOR, "'" + 
			file.toURI().toString().replaceAll("\\\\", "/") + "'");			
		XQResultSequence rseq = expr.executeQuery(q);
		
		while (rseq.next()) {	
			Node node = rseq.getNode();
			String lang = getLanguage(node, defaultLang);
			List<String> values = result.get(lang);
			if (values == null) {
				result.put(lang, values = new ArrayList<String>());
			}
			
			// replace tabs, new line chars, multiple spaces with a single 
			// space
			values.add(node.getTextContent().replaceAll(
				"\\s+", " "));			
		}
		
		rseq.close();
		
		return result;		
	}

	private String getLanguage(Node aNode, String defaultLang)
	{	
		Node node = aNode;
		
		if (node == null) {
			return defaultLang;
		}
		
		NamedNodeMap map = node.getAttributes();
		if (map == null) {
			return defaultLang;
		}
		
		while (node != null) {
			Node lang = map.getNamedItemNS(XMLConstants.XML_NS_URI, XML_LANG);
			if (lang != null) {
				return lang.getTextContent();
			}
			node = node.getParentNode();
		}
		
		return defaultLang;
	}
}
