package pl.psnc.dlibra.app.extension.xml;

import java.awt.Window;
import java.awt.event.ActionEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.ResourceBundle;

import pl.psnc.dlibra.app.common.component.extension.AbstractConfigurableExtension;
import pl.psnc.dlibra.app.extension.metadata.MetadataFinder;
import pl.psnc.dlibra.app.extension.metadata.MetadataFinderException;
import pl.psnc.dlibra.app.extension.metadata.MetadataFinderException.Operation;

/**
 * Finds metadata in XML files basis on given configuration. The configuration
 * is composed of two files:
 * 1. tests.properties - contains XQuery queries which test the type of XML file
 * 2. conversion.properties - contains conversion rules for specific types of
 * XML file.
 * @version $Id: XMLMetadataFinder.java,v 1.9 2009-02-19 14:03:07 tparkola Exp $
 */
public class XMLMetadataFinder
	extends AbstractConfigurableExtension
	implements MetadataFinder
{

	/**
	 * XML conversion file name (contains conversion rules for XML file types).
	 */
	public static final String CONVERSION_RULES = "conversion.properties";

	/**
	 * XML conversion file name (file which contains XML type test).
	 */
	public static final String CONVERSION_TESTS = "tests.properties";

	private static final String XML_FILE_EXTENSION = ".xml";

	private static final String RDF_FILE_EXTENSION = ".rdf";

	private XMLMetadataConverter converter;


	/**
	 * Initializes this component.
	 */
	public XMLMetadataFinder()
	{
		DEFAULT_BUNDLE_PATH = "pl/psnc/dlibra/app/extension/xml/res/GUI";
		DEFAULT_CONFIG_FILES_PACKAGE = "/pl/psnc/dlibra/app/extension/xml/";
		reloadResourceBundle(DEFAULT_BUNDLE_PATH);
	}


	/**
	 * Reloads resources using given path to bundles.
	 * @param bundlePath
	 */
	@Override
	protected void reloadResourceBundle(String bundlePath)
	{
		textsBundle = ResourceBundle.getBundle(bundlePath, Locale.getDefault());
		if (converter != null) {
			converter.setResourceBundle(textsBundle);
		}
	}


	public List<Map<String, Map<String, List<String>>>> loadMetadata(
			Window parentWindow, File file)
		throws UnsupportedOperationException, MetadataFinderException
	{
		if (converter == null) {
			throw new MetadataFinderException(Operation.Load, textsBundle
					.getString("MetadataLoad.convervetNotInitialized.reason"));
		}

		return converter.convert(file);
	}


	public boolean supportsType(FinderType type)
	{
		return FinderType.IMPORT.equals(type);
	}


	@Override
	public String toString()
	{
		if (!Locale.getDefault().equals(textsBundle.getLocale())) {
			reloadResourceBundle(DEFAULT_BUNDLE_PATH);
		}
		return textsBundle.getString("MarcMetadataFinder.name");
	}


	public void storeMetadata(Window parentWindow,
			List<Object> metadataIdentifiers,
			List<Map<String, Map<String, List<String>>>> metadata,
			File destinationFile)
		throws UnsupportedOperationException
	{
		throw new UnsupportedOperationException(
				"Export not supported by the MASTER format metadata finder.");
	}


	public List<String> getDefaultExtensions()
	{
		List<String> exts = new ArrayList<String>(2);
		exts.add(XML_FILE_EXTENSION.substring(1));
		exts.add(RDF_FILE_EXTENSION.substring(1));
		return exts;
	}


	@Override
	public void applyActionPerformed(ActionEvent e)
	{
		super.applyActionPerformed(e);

		List<String> filesNames = new ArrayList<String>();
		filesNames.add(CONVERSION_RULES);
		filesNames.add(CONVERSION_TESTS);

		retrieveConfigurationFiles(e, filesNames);

		Properties tests = load(CONVERSION_TESTS);
		Properties convs = load(CONVERSION_RULES);

		converter = new XMLMetadataConverter(new XMLConversionRules(tests,
				convs));
		converter.setResourceBundle(textsBundle);
	}
}
