package pl.psnc.dlibra.content.conversion;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import pl.psnc.dlibra.plugin.ImageConversionType;

public class QualityImageConverter
	extends CmdExecutor
	implements ImageConverter
{

	/**
	 * Logger for this class
	 */
	private static final Logger logger = Logger
			.getLogger(QualityImageConverter.class);

	private static final int STRIP_COMMAND = 0;

	private static final int RESIZE_COMMAND = 1;

	private static final int DPI_COMMAND = 2;

	private static final int IDENTIFY_COMMAND = 3;

	// attribute rdf name
	private static final String RDF_NAME = "rdfname";

	// value can be a regular expression or explicit string
	private static final String VALUE = "value";

	// max image width
	private static final String WIDTH = "width";

	// max image height
	private static final String HEIGHT = "height";

	// DPI, optional parameter
	private static final String DPI = "dpi";

	private static final String QUALITY_FILENAME_ATTRIBUTE = "qualityFile";

	private List<QualityProperty> qualities = new ArrayList<QualityProperty>();


	private void loadQualities(Properties props)
	{
		int i = 1;

		String rdfName;
		String value;
		int width;
		int height;
		int dpi;

		while (true) {
			rdfName = props.getProperty(prepareName(RDF_NAME, i));
			if (rdfName == null)
				break;

			value = props.getProperty(prepareName(VALUE, i));
			try {
				width = Integer.parseInt(props
						.getProperty(prepareName(WIDTH, i)));
			}
			catch (Throwable e) {
				continue;
			}
			try {
				height = Integer.parseInt(props.getProperty(prepareName(HEIGHT,
					i)));
			}
			catch (Throwable e) {
				continue;
			}
			try {
				dpi = Integer.parseInt(props.getProperty(prepareName(DPI, i)));
				qualities.add(new QualityProperty(rdfName, value, width,
						height, dpi));
			}
			catch (Throwable e) {
				qualities
						.add(new QualityProperty(rdfName, value, width, height));
			}
			i++;
		}
	}


	@Override
	public boolean canConvert(ImageConversionType type, String mime)
	{
		if (type == null || mime == null || mime.isEmpty())
			return false;

		boolean canConvert = false;

		for (String mt : mimeTypes) {
			if (mt.equals(mime)) {
				canConvert = true;
				break;
			}
		}

		return canConvert & type.equals(getType());
	}


	/**
	* Convert file from path using defined commands
	* 
	* @param rdfName rdf name of attribute
	* @param value regex for attribute value
	* @throws IOException 
	*/
	@Override
	public File convert(String rdfName, String value, File toConvert)
		throws IOException
	{
		if (toConvert == null)
			return null;

		if (!toConvert.exists())
			throw new FileNotFoundException("File not found: "
					+ toConvert.getAbsolutePath());

		File outFolder = new File(outputFolder);

		QualityProperty qp = getQualityProperty(rdfName, value);
		if (qp == null)
			return null;

		if (!convertNecessary(toConvert, qp))
			return toConvert;

		File strip = null;
		File resize = null;

		try {
			// strip parameters
			strip = File.createTempFile("convert", "_strip", outFolder);

			List<String> parameters = new ArrayList<String>();
			parameters.add(toConvert.getAbsolutePath());
			parameters.add(strip.getAbsolutePath());

			if (!runCmd(STRIP_COMMAND, parameters).getResult())
				return null;

			// resize command
			resize = File.createTempFile("convert", "_resize", outFolder);

			parameters.clear();
			parameters.add(strip.getAbsolutePath());
			parameters.add(String.valueOf(qp.getWidth()));
			parameters.add(String.valueOf(qp.getHeight()));
			parameters.add(resize.getAbsolutePath());

			if (!runCmd(RESIZE_COMMAND, parameters).getResult())
				return null;

			File result = File.createTempFile("convert", "_result", outFolder);

			parameters.clear();
			parameters.add(resize.getAbsolutePath());
			parameters.add(String.valueOf(qp.getDpi()));
			parameters.add(result.getAbsolutePath());

			if (!runCmd(DPI_COMMAND, parameters).getResult())
				return null;

			return result;
		}
		catch (IOException e) {
			logger.error("Problem while converting.", e);
		}
		finally {
			if (strip != null)
				strip.delete();
			if (resize != null)
				resize.delete();
		}

		return null;

	}


	private boolean convertNecessary(File toConvert, QualityProperty qp)
	{
		// when the resolution is smaller than required converting is unnecessary

		List<String> parameters = new ArrayList<String>();
		parameters.add(toConvert.getAbsolutePath());

		CommandResult result = runCmd(IDENTIFY_COMMAND, parameters);

		// when result is false convert is not necessary as it will probably fail
		if (!result.getResult())
			return false;

		// result is widthxheight
		String[] size = getDimension(result.getStdOut());
		if (size.length != 2)
			return false;

		int width = Integer.valueOf(size[0]).intValue();
		int height = Integer.valueOf(size[1]).intValue();

		if (qp.getWidth() < width || qp.getHeight() < height)
			return true;

		return false;
	}


	/**
	 * Retrive image dimensions by call {@link #IDENTIFY_COMMAND}
	 * @param path Path to image file  
	 * @return Dimension of image or null if some error occurs
	 */
	private String[] getDimension(String output)
	{
		// get size in width:height format
		String[] size = output.split(":");
		if (size.length != 2) {
			return null;
		}

		String width = size[0];
		int pos = 0;

		while (pos < width.length()) {
			try {
				Integer.parseInt((String) width.substring(pos, pos + 1));
				break;
			}
			catch (NumberFormatException e) {
				pos++;
			}
		}
		if (pos > 0) {
			width = width.substring(pos);
			size[0] = width;
		}

		String height = size[1];
		pos = 0;

		while (pos < height.length()) {
			try {
				Integer.parseInt((String) height.substring(pos, pos + 1));
				pos++;
			}
			catch (NumberFormatException e) {
				break;
			}
		}
		if (pos < height.length()) {
			height = height.substring(0, pos);
			size[1] = height;
		}

		return size;
	}


	private QualityProperty getQualityProperty(String rdfName, String value)
	{
		for (QualityProperty p : qualities) {
			if (!p.getRdfName().equals(rdfName))
				continue;

			Pattern pattern = Pattern.compile(p.getValue());
			Matcher matcher = pattern.matcher(value);

			if (matcher.matches())
				return p;
		}
		return null;
	}


	@Override
	protected void doStart()
		throws Exception
	{
		super.doStart();
		loadQualities(loadConfigFile(QUALITY_FILENAME_ATTRIBUTE));
	}


	@Override
	protected void doStop()
		throws Exception
	{
		// nothing to do here.
	}


	@Override
	public Set<String> getRdfNames()
	{
		Set<String> rdfNames = new HashSet<String>();

		for (QualityProperty qp : qualities) {
			rdfNames.add(qp.getRdfName());
		}

		return rdfNames;
	}


	@Override
	public ImageConversionType getType()
	{
		return ImageConversionType.QUALITY;
	}
}
