package pl.psnc.dlibra.content.extraction;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.Logger;

import com.lizardtech.djvu.CachedInputStream;
import com.lizardtech.djvu.DLibraDjVuPage;
import com.lizardtech.djvu.DjVmDir;
import com.lizardtech.djvu.DjVuPage;
import com.lizardtech.djvu.Document;
import com.lizardtech.djvu.text.DjVuText;

/**
 * Class for extracting content of DjVu files.
 * 
 */
public class DjVuTextualContentExtractor
	implements TextualContentExtractor
{

	private static final String IMAGE_X_DJVU = "image/x.djvu";

	/**
	 * Logger for this class
	 */
	private static final Logger logger = Logger
			.getLogger(DjVuTextualContentExtractor.class);


	private String extractTextFromDataPool(DjVuText djvuText, URL url,
			DjVuPage page)
		throws IOException, UnsupportedEncodingException
	{
		CachedInputStream cachedInputStream = CachedInputStream
				.createCachedInputStream(page).init(url, false);
		return extractTextFromStream(djvuText, cachedInputStream);
	}


	private String extractTextFromStream(DjVuText djvuText,
			CachedInputStream cachedInputStream)
		throws IOException, UnsupportedEncodingException
	{
		String result = "";
		DjVuText extractedText = djvuText.init(cachedInputStream);
		if (extractedText != null) {
			extractedText.normalize_text();
			String textString = extractedText.toString();
			if (djvuText.isUTF8) {
				result = textString;
			}
			// Force UTF-8
			byte[] bytes = textString.getBytes();
			result = new String(bytes, "UTF8");
		}
		if (logger.isDebugEnabled()) {
			logger.debug("Text extracted from DjVu file:\n" + result);
		}
		return result;
	}


	public String getTextualContent(File file, String mimeType, String version,
			Integer pageNumber, File mainFile)
		throws FileNotFoundException, IOException
	{
		try {
			StringBuffer content = new StringBuffer();
			URL url = file.toURI().toURL();
			URL mainUrl = mainFile.toURI().toURL();

			try {
				Document djvuDocument = new Document(url);
				DjVmDir dir = djvuDocument.getDjVmDir();
				if (dir.is_indirect()) {
					try {
						DjVuPage page = new DLibraDjVuPage();
						DjVuPage.setIdConverter(new DjvuIdentifierConverterImpl());
						long start = System.currentTimeMillis();
						page.decode(url);
						if (logger.isDebugEnabled()) {
							logger.debug("Decoding page took "
									+ String.valueOf(System.currentTimeMillis()
											- start) + " ms.");
							start = System.currentTimeMillis();
						}
						DjVuText djvuText = DjVuText.createDjVuText(page);
						String text = extractTextFromDataPool(djvuText, url,
							page);
						if (text != null && text.length() > 0) {
							content.append(text);

							content.append("\n");
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_PREFIX);
							content.append(" ");
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_FILE);
							content.append(version);
							content.append(" ");
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_PAGE);

							content.append(String.valueOf(new Document(mainUrl)
									.getPageno("#" + file.getName()) + 1));
							content.append(" ");
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_SUFFIX);
							content.append("\n");
						}
						if (logger.isDebugEnabled()) {
							logger.debug("Extracting text took "
									+ String.valueOf(System.currentTimeMillis()
											- start) + " ms.");
						}
					}
					catch (IllegalStateException e) {
						logger.debug("Exception while processing " + url, e);
						// DjVu index file or some other "utility" file was
						// found
						// skipping it
						return null;
					}
				}
				else {
					Vector v = dir.get_files_list();
					DjVuText djvuText = DjVuText.createDjVuText(djvuDocument);
					for (Iterator iter = v.iterator(); iter.hasNext();) {
						DjVmDir.File djvuFile = (DjVmDir.File) iter.next();
						int pageNum = djvuFile.get_page_num();
						if (pageNum >= 0) {
							CachedInputStream pageStream = djvuDocument
									.get_data(pageNum);
							content.append(extractTextFromStream(djvuText,
								pageStream));

							content.append("\n");
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_PREFIX);
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_FILE);
							content.append(version);
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_PAGE);
							content.append(pageNum);
							content
									.append(TextualContentExtractor.CONTENT_SEPARATOR_SUFFIX);
							content.append("\n");
						}
					}
				}

			}
			catch (ArrayIndexOutOfBoundsException e) {
				if (!"1".equals(e.getMessage())) {
					// known bug with text areas with only one byte length
					throw e;
				}
			}

			if (content.length() > 0) {
				return content.toString();
			}
		}
		catch (Exception e) {
			logger.warn("Exception while indexing DjVU " + file + "("
					+ e.getClass().getName() + "). Message: " + e.getMessage(),
				e);
		}
		return null;
	}


	public boolean isAbleToExtract(String path, String mimeType)
	{
		return IMAGE_X_DJVU.equals(mimeType);
	}
}
