package pl.psnc.dlibra.applet.jpg.viewer;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLDecoder;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.crypto.SecretKey;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import pl.psnc.dlibra.app.extension.fh.jpg.FileInfo;
import pl.psnc.dlibra.app.extension.fh.jpg.ListOfObjectsEditor;
import pl.psnc.dlibra.app.extension.fh.jpg.XMLDescriptionFileParser;
import pl.psnc.dlibra.ps.crypto.KeyAgreementRequestPayload;
import pl.psnc.dlibra.ps.crypto.KeyAgreementResponsePayload;
import pl.psnc.dlibra.ps.crypto.KeyAgreementServer;
import pl.psnc.util.crypto.CryptoUtils;

/**
 * 
 */
public class ConnectionManager
{

	private static final String THUMBNAIL_PATH = "i2j:jpg_thumbnail/";

	private static final String SESSION_ID = "jsessionid";

	private static final String CONTENT_SERVLET = "/Content/";

	private static final String KEY_AGREEMENT_SERVLET = "ka";

	private static final String WATERMARK_SERVLET = "watermark";

	private String serverPath;

	private String contentPath;

	private String mainFilePath;

	private String sessionId;

	private final CryptoUtils cryptoUtils;

	private KeyAgreementRequestPayload karp;


	/**
	 * Creates connection manager and connects to the server to perform
	 * key agreement (may take some time).
	 * @param initialUrl path to the main file of JPG publication (typically, PresentationData.xml)
	 * @param sessionId id of the session
	 * @throws Exception 
	 */
	public ConnectionManager(String initialUrl, String sessionId)
		throws Exception
	{
		int indexOfContent = initialUrl.lastIndexOf(CONTENT_SERVLET);
		serverPath = initialUrl.substring(0, indexOfContent) + "/";
		contentPath = initialUrl.substring(0,
			initialUrl.indexOf("/", indexOfContent + 9)) + "/";
		mainFilePath = initialUrl;
		this.sessionId = sessionId;

		cryptoUtils = getCryptoUtils();
	}


	private CryptoUtils getCryptoUtils()
		throws Exception
	{
		performKeyAgreementPhase1(); //if key existed on server, this will remove it

		// try connection without key, to check if publication is secured
		HttpURLConnection connection = (HttpURLConnection) new URL(
				toUrl(mainFilePath)).openConnection();
		connection.setRequestMethod("HEAD");
		boolean isSecured = (connection.getResponseCode() != HttpURLConnection.HTTP_OK);

		// phase 2 must be always performed, otherwise phase 1 will fail on next run
		SecretKey key = performKeyAgreementPhase2();
		return isSecured ? new CryptoUtils(key) : null;
	}


	public List<FileInfo> downloadFileInfos()
		throws IOException
	{
		String URL = toUrl(mainFilePath);
		InputStream is = getDecipheredInputStream(URL);
		List<FileInfo> fileInfos;
		if (mainFilePath.contains(ListOfObjectsEditor.PRESENTATION_DATA_FILENAME)) {
			try {
				fileInfos = XMLDescriptionFileParser.parse(is);
			}
			catch (SAXException e) {
				throw new IOException("Invalid XML in main file", e);
			}
			catch (ParserConfigurationException e) {
				throw new IOException("Invalid XML in main file", e);
			}			
		}
		else {
			// if it's not PresentationData.xml, it's probably an image
			String filePath = mainFilePath.substring(contentPath.length());
			filePath = new StringTokenizer(filePath, ";?#").nextToken();
			filePath = URLDecoder.decode(filePath, "UTF-8");
			FileInfo fileInfo = new FileInfo(filePath);
			fileInfos = Arrays.asList(fileInfo);
		}
		is.close();
		return fileInfos;
	}


	public Properties downloadWatermarkProperties() throws IOException
	{
		String url = serverPath + WATERMARK_SERVLET;
		url = toUrl(url);
		InputStream is = getDecipheredInputStream(url);
		Properties result = new Properties();
		result.loadFromXML(is);
		is.close();
		return result;
	}


	public InputStream getImageStream(FileInfo fileInfo)
		throws IOException
	{
		try {
			String url = toUrl(contentPath + encode(fileInfo.getFilePath()));
			return getDecipheredInputStream(url);
		}
		catch (UnsupportedEncodingException e) {
			throw new RuntimeException("This shouldn't happen");
		}
	}


	public InputStream getThumbnailStream(FileInfo fileInfo)
		throws IOException
	{
		try {
			String url = toUrl(contentPath + THUMBNAIL_PATH
					+ encode(fileInfo.getFilePath()));
			return getDecipheredInputStream(url);
		}
		catch (UnsupportedEncodingException e) {
			throw new RuntimeException("This shouldn't happen");
		}
	}


	/**
	 * Encodes given path so that it can be used in URLs.
	 */
	private String encode(String filePath)
	{
		try {
			URI uri = new URI(null, null, filePath, null, null);
			return uri.toASCIIString();
		}
		catch (URISyntaxException e) {
			throw new RuntimeException("Invalid path? (" + filePath + ")", e);
		}
	}


	private InputStream getDecipheredInputStream(String url)
		throws IOException
	{
		URLConnection connection = new URL(url).openConnection();
		connection.setRequestProperty("User-Agent", "dLibra-html-plugin/"
				+ VersionProvider.getVersion() + " (http://dlibra.psnc.pl/)");
		if (cryptoUtils != null) {
			return cryptoUtils.getDecipheredInputStream(connection
					.getInputStream());
		}
		return connection.getInputStream();
	}


	public InputStream getInputStream(String url)
		throws IOException
	{
		if (!url.startsWith("http://")) {
			url = serverPath + url;
		}
		URLConnection connection = new URL(url).openConnection();
		connection.setRequestProperty("User-Agent", "dLibra-html-plugin/"
				+ VersionProvider.getVersion() + " (http://dlibra.psnc.pl/)");
		return connection.getInputStream();
	}


	private String toUrl(String url)
	{
		if (url.indexOf(";") < 0) {
			if (url.indexOf("#") > -1) {
				url = url.substring(0, url.indexOf("#"));
			}
			if (url.indexOf('?') < 0) {
				return url + ";" + SESSION_ID + "=" + sessionId;
			}
			return url.substring(0, url.indexOf('?')) + ";" + SESSION_ID
					+ "=" + sessionId + url.substring(url.indexOf('?'));
		}
		return url;
	}


	private void performKeyAgreementPhase1()
		throws Exception
	{
		// Send the first request
		String targetUrl = serverPath + KEY_AGREEMENT_SERVLET + ";"
				+ SESSION_ID + "=" + sessionId;
		URL url = new URL(targetUrl);

		HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		conn.setRequestMethod("POST");
		conn.setAllowUserInteraction(false);
		conn.setDoOutput(true);
		conn.setRequestProperty("Content-type", "binary/x-java-serialized");
		conn.getOutputStream().close();

		// Get the response
		conn.connect();

		InputStream is = conn.getInputStream();
		ObjectInputStream ois = null;
		ois = new ObjectInputStream(is);

		karp = (KeyAgreementRequestPayload) ois.readObject();
		ois.close();
	}


	private SecretKey performKeyAgreementPhase2()
		throws Exception
	{
		String targetUrl = serverPath + KEY_AGREEMENT_SERVLET + ";"
				+ SESSION_ID + "=" + sessionId;
		URL url = new URL(targetUrl);

		// Do the first phase
		KeyAgreementServer kas = new KeyAgreementServer();
		KeyAgreementResponsePayload karep;

		karep = kas.process(karp);

		// Send the results to the client
		HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		conn.setRequestMethod("POST");
		conn.setAllowUserInteraction(false);
		conn.setDoOutput(true);
		conn.setRequestProperty("Content-type", "binary/x-java-serialized");

		ObjectOutputStream oos = new ObjectOutputStream(conn.getOutputStream());
		oos.writeObject(karep);
		oos.flush();
		oos.close();

		// We don't read any input here
		conn.connect();
		conn.getInputStream().close();

		// Do the second phase
		return kas.finish();
	}


	public boolean isSecured()
	{
		return cryptoUtils != null;
	}
}
