/*
 * Copyright (c) 2011 Poznan Supercomputing and Networking Center
 * 10 Noskowskiego Street, Poznan, Wielkopolska 61-704, Poland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Poznan Supercomputing and Networking Center ("Confidential Information").
 * You shall not disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered into
 * with PSNC.
 */
package pl.psnc.dlibra.applet.jpg.viewer;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.text.AttributedString;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.imageio.ImageIO;
import javax.swing.JPanel;


/**
 * Creates watermark image and applies it to shown images.
 */
public class Watermark
{
	private static enum Align { LEFT, CENTER, RIGHT };
	
	private static class TextInfo
	{
	
		public String text;
	
		public Font[] fonts;

		public Color color;
	
		public int x, y;
	
		public int width, height;

		public Align align;
	
		public boolean wrap;
	}


	private static final String CONFIG_IMAGE_PATH = "image.path";
	private static final String CONFIG_OPACITY = "opacity";
	private static final String CONFIG_PREFIX_TEXT = "text.";
	private static final String CONFIG_SUFIX_POSITION = ".position";
	private static final String CONFIG_SUFIX_SIZE = ".size";
	private static final String CONFIG_SUFIX_ALIGN = ".align";
	private static final String CONFIG_SUFIX_FONT = ".font";
	private static final String CONFIG_SUFIX_FONT_SIZE = ".font.size";
	private static final String CONFIG_SUFIX_FONT_STYLE = ".font.style";
	private static final String CONFIG_SUFIX_FONT_COLOR = ".font.color";
	private static final String CONFIG_SUFIX_WORDWRAP= ".wordwrap";

	private BufferedImage graphic;

	private float opacity;

	public void init(Properties config, ConnectionManager cm)
		throws IOException
	{
		String imagePath = config.getProperty(CONFIG_IMAGE_PATH);
		if (imagePath != null) {
			InputStream is = cm.getInputStream(imagePath);
			graphic = ImageIO.read(is);
			is.close();
		}
		else {
			graphic = new BufferedImage(500, 500, BufferedImage.TYPE_INT_ARGB);
		}

		String opacity = config.getProperty(CONFIG_OPACITY);
		try {
			this.opacity = Float.valueOf(opacity);
		}
		catch (Exception e) {
			this.opacity = 0.5f;
		}

		List<TextInfo> textInfos = loadTextInfos(config);
		Graphics2D g2d = graphic.createGraphics();
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			RenderingHints.VALUE_ANTIALIAS_ON);
		for (TextInfo textInfo : textInfos) {
			drawText(textInfo, g2d);
		}
	}


	public BufferedImage apply(BufferedImage image)
	{
		final int ww = graphic.getWidth();
		final int wh = graphic.getHeight();
		final int iw = image.getWidth();
		final int ih = image.getHeight();
		double scale = Math.min(1.0 * iw / ww, 1.0 * ih / wh);
		int ww2 = (int) (ww * scale);
		int wh2 = (int) (wh * scale);

		//create new image in case the original one doesn't have enough color depth
		BufferedImage applied = new BufferedImage(iw, ih, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2d = (Graphics2D) applied.getGraphics();
		g2d.drawImage(image, 0, 0, null);

		g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
			RenderingHints.VALUE_INTERPOLATION_BILINEAR);
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			RenderingHints.VALUE_ANTIALIAS_ON);
		g2d.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
			RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
		AlphaComposite composite = AlphaComposite.SrcOver.derive(opacity);
		g2d.setComposite(composite);
		g2d.drawImage(graphic, (iw - ww2) / 2, (ih - wh2) / 2, ww2, wh2, null);

		return applied;
	}


	private List<TextInfo> loadTextInfos(Properties config)
	{
		ArrayList<TextInfo> textInfos = new ArrayList<TextInfo>();
		for (Object key : config.keySet()) {
			String keyString = (String) key;
			if (keyString.startsWith(CONFIG_PREFIX_TEXT)
					&& keyString.lastIndexOf('.') == CONFIG_PREFIX_TEXT
							.lastIndexOf('.')) {
				TextInfo textInfo = new TextInfo();
				textInfo.text = config.getProperty(keyString);

				String position = config.getProperty(keyString + CONFIG_SUFIX_POSITION);
				try {
					StringTokenizer tokenizer = new StringTokenizer(position, ", ");
					textInfo.x = Integer.valueOf(tokenizer.nextToken());
					textInfo.y = Integer.valueOf(tokenizer.nextToken());
				}
				catch (Exception e) {
					textInfo.x = graphic.getWidth() / 2;
					textInfo.y = graphic.getHeight() / 2;
				}

				String size = config.getProperty(keyString + CONFIG_SUFIX_SIZE);
				try {
					StringTokenizer tokenizer = new StringTokenizer(size, ", ");
					textInfo.width = Integer.valueOf(tokenizer.nextToken());
					textInfo.height = Integer.valueOf(tokenizer.nextToken());
				}
				catch (Exception e) {
					textInfo.width = textInfo.height = 0;
				}

				String align = config.getProperty(keyString + CONFIG_SUFIX_ALIGN);
				try {
					textInfo.align = Align.valueOf(align.toUpperCase());
				}
				catch (Exception e) {
					textInfo.align = Align.LEFT;
				}

				Font baseFont;
				String font = config.getProperty(keyString + CONFIG_SUFIX_FONT);
				if (font != null) {
					baseFont = new Font(font, Font.PLAIN, 10);
				}
				else {
					baseFont = new JPanel().getFont();
				}
				String fontStyle = config.getProperty(keyString + CONFIG_SUFIX_FONT_STYLE);
				if (fontStyle != null) {
					fontStyle = fontStyle.toLowerCase();
					int style = 0;
					if (fontStyle.contains("b")) {
						style += Font.BOLD;
					}
					if (fontStyle.contains("i")) {
						style += Font.ITALIC;
					}
					baseFont = baseFont.deriveFont(style);
				}
				String fontSize = config.getProperty(keyString + CONFIG_SUFIX_FONT_SIZE);
				try {
					float fsize = Float.valueOf(fontSize);
					baseFont = baseFont.deriveFont(fsize);
					textInfo.fonts = new Font[] { baseFont};
				}
				catch (Exception e) {
					textInfo.fonts = prepareFontCache(baseFont);
				}
				String fontColor = config.getProperty(keyString + CONFIG_SUFIX_FONT_COLOR);
				try {
					textInfo.color = Color.decode(fontColor);
				}
				catch (Exception e) {
					textInfo.color = Color.black;
				}

				String wrap = config.getProperty(keyString + CONFIG_SUFIX_WORDWRAP);
				textInfo.wrap = Boolean.TRUE.toString().equals(wrap);

				textInfos.add(textInfo);
			}
		}
		return textInfos;
	}


	private Font[] prepareFontCache(Font baseFont)
	{
		final int[] sizeDiffIntervals = { 5, 15, 64, 257};
		ArrayList<Font> fonts = new ArrayList<Font>();
		int sizeDiff = 1;
		for (int i = 0; i < sizeDiffIntervals.length - 1; i++) {
			for (int size = sizeDiffIntervals[i]; size < sizeDiffIntervals[i + 1]; size += sizeDiff) {
				fonts.add(baseFont.deriveFont((float) size));
			}
			sizeDiff *= 2;
		}
		return (Font[]) fonts.toArray(new Font[fonts.size()]);
	}


	protected void drawText(TextInfo textInfo, Graphics2D g2d)
	{
		LineBreakMeasurer lineBreaker = findBestLineBreaker(textInfo, g2d);
		lineBreaker.setPosition(0);
		g2d.setColor(textInfo.color);

		final float bestWidth = textInfo.width > 0 ? textInfo.width : graphic
				.getWidth();
		final float bestHeight = textInfo.height > 0 ? textInfo.height
				: graphic.getHeight();
		final String label = textInfo.text;
		float y = (float) textInfo.y;
		while (lineBreaker.getPosition() < label.length()) {
			TextLayout textLayout = lineBreaker.nextLayout(bestWidth,
				label.length(), true);
			if (textLayout == null) {
				textLayout = lineBreaker.nextLayout(bestWidth, label.length(),
					false);
			}
			y += textLayout.getAscent();
			if (y > textInfo.y + bestHeight) {
				break;
			}
			float x;
			switch (textInfo.align) {
				case RIGHT:
					if (textInfo.width > 0) {
						x = textInfo.x + textInfo.width
								- textLayout.getAdvance();
					}
					else {
						x = textInfo.x - textLayout.getAdvance();
					}
					break;
				case CENTER:
					if (textInfo.width > 0) {
						x = textInfo.x
								+ (textInfo.width - textLayout.getAdvance())
								/ 2;
					}
					else {
						x = textInfo.x - textLayout.getAdvance() / 2;
					}
					break;
				default:
					x = textInfo.x;
			}
			textLayout.draw(g2d, x, y);
			y += textLayout.getDescent() + textLayout.getLeading();
		}
	}


	private LineBreakMeasurer findBestLineBreaker(TextInfo textInfo,
			Graphics2D g)
	{
		final Font[] fontCache = textInfo.fonts;
		final String text = textInfo.text;
		final int width = textInfo.width;
		final int height = textInfo.height;

		LineBreakMeasurer bestLineBreaker = null;
		int lo = 0;
		int hi = fontCache.length - 1;
		while (hi - lo > 1) {
			int mid = (hi + lo) / 2;

			LineBreakMeasurer linebreaker = createLineBreaker(text,
				fontCache[mid], g);
			boolean fits = true;
			if (!textInfo.wrap) {
				fits = linebreaker.nextOffset(width) == text.length();
			}
			if (fits) {
				int y = 0;
				while (linebreaker.getPosition() < text.length()) {
					TextLayout textLayout = linebreaker.nextLayout(width,
						text.length(), true);
					if (textLayout == null || textLayout.getAdvance() > width) {
						// one word is longer than whole line, won't fit
						y = height + 1;
						break;
					}
					y += textLayout.getAscent() + textLayout.getDescent()
							+ textLayout.getLeading();
				}
				fits = y <= height;
			}

			if (fits) {
				lo = mid;
				bestLineBreaker = linebreaker;
			}
			else {
				hi = mid;
			}
		}
		if (bestLineBreaker == null) {
			bestLineBreaker = createLineBreaker(text, fontCache[0], g);
		}
		return bestLineBreaker;
	}


	private LineBreakMeasurer createLineBreaker(String label, Font font,
			Graphics2D g)
	{
		HashMap<TextAttribute, Object> attribs = new HashMap<TextAttribute, Object>();
		attribs.put(TextAttribute.FONT, font);
		return new LineBreakMeasurer(
				new AttributedString(label, attribs).getIterator(),
				g.getFontRenderContext());
	}
}
